/*
 * in_sox.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"
#define __DYNAMIC__
#include "libsox_stub.h"

struct track {
  samplebuf_t *buf;
  sox_format_t *in;
  sox_effects_chain_t *chain;

  HANDLE hThreadSoX;
};

static track_t track_play;
static track_t track_info;

/* The function that will be called to input samples into the effects chain.
 * In this example, we get samples to process from a SoX-openned audio file.
 * In a different application, they might be generated or come from a different
 * part of the application. */
static int input_drain(
    sox_effect_t * effp, sox_sample_t * obuf, size_t * osamp)
{
  //(void)effp;   /* This parameter is not needed in this example */
  //sox_format_t *in=*(sox_format_t **)effp->priv;
  track_t *track=*(track_t **)effp->priv;
  sox_format_t *in=track->in;

  /* ensure that *osamp is a multiple of the number of channels. */
  *osamp -= *osamp % effp->out_signal.channels;

  /* Read up to *osamp samples into obuf; store the actual number read
   * back to *osamp */
  *osamp = sox_read(in, obuf, *osamp);

  /* sox_read may return a number that is less than was requested; only if
   * 0 samples is returned does it indicate that end-of-file has been reached
   * or an error has occurred */
  if (!*osamp && in->sox_errno)
    fprintf(stderr, "%s: %s\n", in->filename, in->sox_errstr);

  return *osamp? SOX_SUCCESS : SOX_EOF;
}

/* A `stub' effect handler to handle inputting samples to the effects
 * chain; the only function needed for this example is `drain' */
static sox_effect_handler_t const * input_handler(void)
{
  static sox_effect_handler_t handler = {
    "input", NULL, SOX_EFF_MCHAN, NULL, NULL, NULL, input_drain, NULL, NULL, sizeof(sox_format_t *)
  };
  return &handler;
}

static DWORD WINAPI in_sox_thread(LPVOID p)
{
  track_t *in_sox=p;

  sox_flow_effects(in_sox->chain, NULL, NULL);
  samplebuf_signal_done(in_sox->buf);

  return 0;
}

static track_t *in_sox_open(track_t *track, samplebuf_t *buf, const char *fn)
{
  sox_format_t *in;
  int length;
  const char *tag;
  sox_comments_t comments;

  memset(track, 0, sizeof *track);

  if (NULL==(in=sox_open_read(fn, NULL, NULL, NULL)))
	return NULL;

  track->in=in;
  length=(in->signal.length*10)/(in->signal.rate/100*in->signal.channels); 

  samplebuf_open(buf, &sox_plugin, track, fn, track->in->signal.channels,
      in->signal.rate, in->signal.precision, length);
  track->buf=buf;

  comments=in->oob.comments;
#define MAKE_INFO(id) \
  if (NULL!=(tag=sox_find_comment(comments,#id))) \
    _snprintf (buf->tag.id, sizeof buf->tag.id, "%s", tag);
#define MAKE_ALTERNATE_INFO(id1, id2) \
  if (0==buf->tag.id1[0]&&NULL!=(tag=sox_find_comment(comments,#id2))) \
    _snprintf (buf->tag.id1, sizeof buf->tag.id1, "%s", tag);
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_INFO
  _snprintf(buf->formatinformation,
      sizeof buf->formatinformation,
      "Length: %d seconds"
      "\nDecoder: SoX"
      "\n\nAudio Codec: %s"
      "\nChannels: %d"
      "\nBits per sample (in): %d"
      "\nSample Rate (in): %.f Hz"
      "\nBits per sample (out): %d"
      "\nSample Rate (out): %.f Hz"
      //"\nFile Size: %d bytes"
      //"\nAverage bitrate: %d kbps"
      //"\nCompression Ratio: %d%%"
      ,
      buf->length/1000,
      in->filetype,
      in->signal.channels,
      in->signal.precision,
      (double)in->signal.rate,
      track->buf->bps,
      (double)track->buf->rate);

  return track;
}

static track_t *in_sox_start(track_t *track, vout_t *vout, int (*gettime)(),
    void (*settime)(int), void (*showvis)())
{
  samplebuf_start(track->buf);

  track->chain=track_init_sox_chain(track, track->buf,
      track->in->signal.precision, track->in->signal.rate,
	  track->in->signal.channels, input_handler);

  if (ini.winamp.vis)
    showvis();

  track->hThreadSoX=CreateThread(
    NULL,           // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,              // dwStackSize
	in_sox_thread,  // lpStartAddress
	track,          // lpParameter
	0,              // dwCreationFlags
	NULL            // lpThreadId
  );

  return track;
}

static void in_sox_close(track_t *track)
{
  samplebuf_close(track->buf);

  if (NULL!=track->in)
    sox_close(track->in);

  memset(track, 0, sizeof *track);
}

static void in_sox_stop(track_t *track)
{
  if (NULL!=track->hThreadSoX) {
    WaitForSingleObject(track->hThreadSoX, INFINITE);
    CloseHandle(track->hThreadSoX);
  }

  if (NULL!=track->chain)
    sox_delete_effects_chain(track->chain);

  samplebuf_stop(track->buf);
}

static void in_sox_kill(track_t *track)
{
  samplebuf_signal_done(track->buf);
}

static void in_sox_pause(track_t *track)
  {}
static void in_sox_unpause(track_t *track)
  {}

static track_t *in_sox_create_play(samplebuf_t *buf, const char *fn)
{
#ifndef __XXX__
  track_t *track=in_sox_open(&track_play, buf, fn);

  return NULL==track?ffsox_plugin.create_play(buf, fn):track;
#else // __XXX__
#endif // __XXX__
}

static track_t *in_sox_create_info(samplebuf_t *buf, const char *fn)
{
#ifndef __XXX__
  track_t *track=in_sox_open(&track_info, buf, fn);

  return NULL==track?ffsox_plugin.create_info(buf, fn):track;
#else // __XXX__
#endif // __XXX__
}

static int in_sox_seek(track_t *track, int ms) {
  return 0;
}

static void in_sox_flush(track_t *track) {
}

static void in_sox_restart(track_t *track) {
}

plugin_t sox_plugin = {
  .create_play=in_sox_create_play,
  .create_info=in_sox_create_info,
  .close=in_sox_close,
  .start=in_sox_start,
  .stop=in_sox_stop,
  .kill=in_sox_kill,
  .pause=in_sox_pause,
  .unpause=in_sox_unpause,
  .seek=in_sox_seek,
  .flush=in_sox_flush,
  .restart=in_sox_restart
};
