/*
 * packet_queue.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
//#include <windows.h>
//#include <libavformat/avformat.h>
#include "plugin.h"

static AVPacket *packet_queue_get(packet_queue_t *q, AVPacket *pkt);

packet_queue_t *packet_queue_open(packet_queue_t *q, int min, int max,
    HANDLE hEventWrite)
{
  q->state=PACKET_QUEUE_WORK;
  q->min=min;
  q->max=max;
  q->cur=0;
  q->hEventWrite=hEventWrite;

  q->hEventRead=CreateEvent(
    NULL,           // lpEventAttributes
    FALSE,          // bManualReset
    FALSE,          // bInitialState
    NULL            // lpName
  );

  InitializeCriticalSection(&q->cs);

  return q;
}

void packet_queue_reset(packet_queue_t *q)
{
  AVPacket pkt;

  while (NULL!=q->head)
	av_free_packet(packet_queue_get(q, &pkt));

  q->state=PACKET_QUEUE_WORK;
}

void packet_queue_close(packet_queue_t *q)
{
  packet_queue_reset(q);
  DeleteCriticalSection(&q->cs);
  CloseHandle(q->hEventRead);
}

int packet_queue_overflow(packet_queue_t *q)
  { return PACKET_QUEUE_KILL!=q->state&&q->max<q->cur; }
int packet_queue_underflow(packet_queue_t *q)
  { return q->cur<q->min; }

static void packet_queue_add(packet_queue_t *q, AVPacket *pkt)
{
  AVPacketList *n=av_malloc(sizeof *n);

  av_dup_packet(pkt);
  n->pkt=*pkt;
  n->next=NULL;

  if (q->tail)
    q->tail->next=n;
  else
	q->head=n;

  q->tail=n;
  ++q->cur;
}

static AVPacket *packet_queue_get(packet_queue_t *q, AVPacket *pkt)
{
  AVPacketList *n=q->head;

  if (NULL==(q->head=n->next))
    q->tail=NULL;

  --q->cur;  
  *pkt=n->pkt;
  av_free(n);

  return pkt;
}

void packet_queue_signal_eof(packet_queue_t *q)
{
  if (PACKET_QUEUE_KILL!=q->state) {
    q->state=PACKET_QUEUE_EOF;
    SetEvent(q->hEventRead);
  }
}

void packet_queue_signal_work(packet_queue_t *q)
{
  q->state=PACKET_QUEUE_WORK;
  SetEvent(q->hEventRead);
}

void packet_queue_signal_kill(packet_queue_t *q)
{
  q->state=PACKET_QUEUE_KILL;
  SetEvent(q->hEventRead);
  SetEvent(q->hEventWrite);
}

void packet_queue_write(packet_queue_t *q, AVPacket *pkt)
{
  if (PACKET_QUEUE_KILL!=q->state) {
    EnterCriticalSection(&q->cs);
    packet_queue_add(q, pkt);
    LeaveCriticalSection(&q->cs);
    SetEvent(q->hEventRead);
  }
}

AVPacket *packet_queue_read(packet_queue_t *q, AVPacket *pkt)
{
  EnterCriticalSection(&q->cs);

  while (PACKET_QUEUE_PAUSE==q->state
      ||(NULL==q->head&&PACKET_QUEUE_WORK==q->state)) {

    LeaveCriticalSection(&q->cs);
    WaitForSingleObject(q->hEventRead, INFINITE);
    EnterCriticalSection(&q->cs);
  }

  pkt=NULL==q->head||PACKET_QUEUE_KILL==q->state
	  ?NULL
      :packet_queue_get(q, pkt)
	  ;
  LeaveCriticalSection(&q->cs);

  if (packet_queue_underflow(q))
    SetEvent(q->hEventWrite);

  return pkt;
}
