/*
 * sox.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

/* The function that will be called to output samples from the effects chain.
 * In this example, we store the samples in a SoX-opened audio file.
 * In a different application, they might perhaps be analysed in some way,
 * or displayed as a wave-form */
static int output_flow(sox_effect_t *effp UNUSED, sox_sample_t const * ibuf,
    sox_sample_t * obuf UNUSED, size_t * isamp, size_t * osamp)
{
  /* Write out *isamp samples */
  samplebuf_t *buf=*(samplebuf_t **)effp->priv;
  size_t len = 0;


  if (SAMPLEBUF_DONE==samplebuf_wait_write(buf))
    return SOX_EOF;

  for (;;) {
    if (len<*isamp&&buf->wp<buf->mp) {
      SOX_SAMPLE_LOCALS;
      int32_t sample;
      char *rp=(char *)&sample;

      switch (buf->bps) {
        case 16:
          sample=SOX_SAMPLE_TO_SIGNED_16BIT(*ibuf++,effp->clips);
          *buf->wp++=*rp++;
          *buf->wp++=*rp++;
          break;
        case 24:
          sample=SOX_SAMPLE_TO_SIGNED_24BIT(*ibuf++,effp->clips);
          *buf->wp++=*rp++;
          *buf->wp++=*rp++;
          *buf->wp++=*rp++;
          break;
        case 32:
          sample=SOX_SAMPLE_TO_SIGNED_32BIT(*ibuf++,effp->clips);
          *buf->wp++=*rp++;
          *buf->wp++=*rp++;
          *buf->wp++=*rp++;
          *buf->wp++=*rp++;
          break;
        default:
          return SOX_EOF;
      }

      ++len;
    }
    else {
      if (buf->wp==buf->mp)
        samplebuf_signal_read(buf);

      if (len==*isamp)
        break;
      else if (SAMPLEBUF_DONE==samplebuf_wait_write(buf))
        return SOX_EOF;
    }
  }

  /* len is the number of samples that were actually written out; if this is
   * different to *isamp, then something has gone wrong--most often, it's
   * out of disc space */
  if (len != *isamp) {
    //fprintf(stderr, "%s: %s\n", out->filename, out->sox_errstr);
    return SOX_EOF;
  }

  /* Outputting is the last `effect' in the effect chain so always passes
   * 0 samples on to the next effect (as there isn't one!) */
  *osamp = 0;

  (void)effp;   /* This parameter is not needed in this example */

  return SOX_SUCCESS; /* All samples output successfully */
}

/* A `stub' effect handler to handle outputting samples from the effects
 * chain; the only function needed for this example is `flow' */
static sox_effect_handler_t const * output_handler(void)
{
  static sox_effect_handler_t handler = {
    "output", NULL, SOX_EFF_MCHAN, NULL, NULL, output_flow, NULL, NULL, NULL, sizeof(samplebuf_t *)
  };
  return &handler;
}

static const sox_signalinfo_t raw_signal = {
  44100,                // sox_rate_t       rate;
  2,                    // unsigned         channels;
  16,                   // unsigned         precision;
  0,                    // size_t           length;
  NULL                  // double           * mult;
};

sox_effects_chain_t *track_init_sox_chain(track_t *track,
    samplebuf_t *buf, int in_bps, int in_rate, int in_channels,
	sox_effect_handler_t const *(*input_handler)(void))
{
  sox_effects_chain_t *chain=NULL;
  int out_bps=buf->bps;
  int out_rate=buf->rate;
  int out_channels=buf->nch;
  int dither=out_bps<in_bps?1:0;
  char *gain_val=
      1==ini.rg.mode?buf->tag.replaygain_track_gain:
      2==ini.rg.mode?buf->tag.replaygain_album_gain:
	  NULL;

  sox_signalinfo_t in_signal=raw_signal, tmp_signal, out_signal;
  sox_effect_t * e;

  // SoX
  /* Open the output file; we must specify the output signal characteristics.
   * Since we are using only simple effects, they are the same as the input
   * file characteristics */
  in_signal.precision=in_bps;
  in_signal.rate=in_rate;
  in_signal.channels=in_channels;

  tmp_signal=in_signal;

  out_signal=in_signal;
  out_signal.rate=out_rate;
  out_signal.channels=out_channels;
  out_signal.precision=out_bps;
  //out_signal.precision=ffsox_get_targt_bps(f);
  //out_signal.channels=ffsox_get_targt_nch(f);

  /* Create an effects chain; some effects need to know about the input
   * or output file encoding so we provide that information here */
  chain = sox_create_effects_chain(NULL, NULL);

  if (NULL==chain)
    goto fail;

  /* The first effect in the effect chain must be something that can source
   * samples; in this case, we have defined an input handler that inputs
   * data from an audio file */
  e = sox_create_effect(input_handler());
  *(track_t **)e->priv=track;

  /* This becomes the first `effect' in the chain */
  if (SOX_SUCCESS!=sox_add_effect(chain, e, &tmp_signal, &tmp_signal))
    goto fail;

  if (out_rate!=in_rate) {
    char rate[32];
    char *args[] = { ini.sox.quality, rate };
	int offs=0==**args?1:0;
    /* Create the `dither' effect, and initialise it with the desired parameters: */
    e=sox_create_effect(sox_find_effect("rate"));
    sprintf(rate, "%d", out_rate);
  
    if (SOX_SUCCESS != sox_effect_options(e, 2-offs, args+offs))
      goto fail;

    /* Add the effect to the end of the effects processing chain: */
    if (SOX_SUCCESS!=sox_add_effect(chain,e,&tmp_signal,&tmp_signal))
      goto fail;

    dither=1;
  }

  if (NULL!=gain_val&&*gain_val) {
    double amp=ini.rg.preamp+atof(gain_val);
    char val[32];

    if (0==strcmp(ini.sox.amplify,VOL)) {
      e = sox_create_effect(sox_find_effect("vol"));
      sprintf(val, "%.2fdB", amp);

      if ('-'==val[0]||!ini.sox.limit_switch) {
        char *args[] = { val };

        if (SOX_SUCCESS!=sox_effect_options(e, 1, args))
          goto fail;
	  }
	  else {
        char *args[] = { val, ini.sox.limit };

        if (SOX_SUCCESS!=sox_effect_options(e, 2, args))
          goto fail;
	  }
    }
    else {
      e = sox_create_effect(sox_find_effect("gain"));
      sprintf(val, "%.2f", amp);

      if ('-'==val[0]||!ini.sox.limit_switch) {
        char *args[] = { val };

        if (SOX_SUCCESS!=sox_effect_options(e, 1, args))
          goto fail;
      }
      else {
        char *args[] = { "-l", val };

        if (SOX_SUCCESS!=sox_effect_options(e, 2, args))
          goto fail;
      }
    }

    /* Add the effect to the end of the effects processing chain: */
    if (SOX_SUCCESS!=sox_add_effect(chain, e, &tmp_signal, &tmp_signal))
      goto fail;

    dither=1;
  }

  if (dither&&ini.sox.dither) {
    char *args[] = { "-s" };
    /* Create the `dither' effect, and initialise it with the desired parameters: */
    e=sox_create_effect(sox_find_effect("dither"));
  
    if (SOX_SUCCESS != sox_effect_options(e,ini.sox.noise_shaping?1:0, args))
      goto fail;

    /* Add the effect to the end of the effects processing chain: */
    if (SOX_SUCCESS!=sox_add_effect(chain,e,&tmp_signal,&out_signal))
      goto fail;
  }

  /* The last effect in the effect chain must be something that only consumes
   * samples; in this case, we have defined an output handler that outputs
   * data to an audio file */
  e = sox_create_effect(output_handler());
  *(samplebuf_t **)e->priv=buf;

  if (SOX_SUCCESS != sox_add_effect(chain, e, &tmp_signal, &out_signal))
      goto fail;

  return chain;
fail:
  if (NULL!=chain)
    sox_delete_effects_chain(chain);

  return NULL;
}

