/*
 * timer.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

timer_t *timer_open(timer_t *t, UINT uPeriod)
{
  TIMECAPS tc;

#ifdef __MUTEX__
  if (NULL==(t->hMutex=CreateMutex(NULL, FALSE, NULL)))
    goto fail;
#else
  InitializeCriticalSection(&t->cs);
#endif

  if (NULL==(t->hEvent=CreateEvent(NULL, FALSE, FALSE, NULL)))
    goto fail;

  if (TIMERR_NOERROR!=timeGetDevCaps(&tc, sizeof tc))
    goto fail;

  t->wTimerRes=min(max(tc.wPeriodMin, uPeriod),tc.wPeriodMax);
  timeBeginPeriod(t->wTimerRes);

  return t;
fail:
  return NULL;
}

void timer_close(timer_t *t)
{
  if (0<t->wTimerRes)
    timeEndPeriod(t->wTimerRes);

  if (NULL!=t->hEvent)
    CloseHandle(t->hEvent);

#ifdef __MUTEX__
  if (NULL!=t->hMutex)
    CloseHandle(t->hMutex);
#else
  DeleteCriticalSection(&t->cs);
#endif
}

static void CALLBACK OneShotCallback(UINT wTimerID, UINT msg, DWORD dwUser,
    DWORD dw1, DWORD dw2) 
{ 
  timer_t *t=(timer_t *)dwUser;
#ifdef __MUTEX__
  WaitForSingleObject(t->hMutex, INFINITE);
#else
  EnterCriticalSection(&t->cs);
#endif
  t->wTimerID=0;			// invalidate timer ID (no longer in use)
  SetEvent(t->hEvent);
#ifdef __MUTEX__
  ReleaseMutex(t->hMutex);
#else
  LeaveCriticalSection(&t->cs);
#endif
}

void timer_kill(timer_t *t)
{
#ifdef __MUTEX__
  WaitForSingleObject(t->hMutex, INFINITE);
#else
  EnterCriticalSection(&t->cs);
#endif
  if (t->wTimerID) {
    timeKillEvent(t->wTimerID);
    t->wTimerID=0;			// invalidate timer ID (no longer in use)
    SetEvent(t->hEvent);
  }
#ifdef __MUTEX__
  ReleaseMutex(t->hMutex);
#else
  LeaveCriticalSection(&t->cs);
#endif
}

void timer_sleep(timer_t *t, UINT msInterval)
{
#ifdef __MUTEX__
  WaitForSingleObject(t->hMutex, INFINITE);
#else
  EnterCriticalSection(&t->cs);
#endif
  t->wTimerID = timeSetEvent(
      msInterval,			// delay
      t->wTimerRes,			// resolution (global variable)
      OneShotCallback,		// callback function
      (DWORD)t,				// user data
      TIME_ONESHOT);		// single timer event
#ifdef __MUTEX__
  ReleaseMutex(t->hMutex);
#else
  LeaveCriticalSection(&t->cs);
#endif

  while (t->wTimerID)
    WaitForSingleObject(t->hEvent, INFINITE);
}
